#include "clm.h"
#include "clm_gpu.h"

#include <stdio.h>

struct clm_NativeBuf {};

int clm_gpuInit(unsigned int mode) {
	return 0;
}

void clm_gpuDestroy() {}

void clm_linearInit(clm_Linear *linear) {}

void clm_linearForward(clm_Linear *linear, unsigned int batchSize, clm_Matrix *inputs) {
	clm_Matrix *outputs = linear->output.matrixes;
	for(unsigned int b = 0; b < batchSize; b++) {
		clm_matrixMultiplyMatrix(linear->weights.matrix, inputs[b], outputs[b]);

		if(clm_matrixIsInvalid(outputs[b])) {
			printf("Forward pass failed\n");
			return;
		}

		clm_matrixAddMatrix(outputs[b], linear->bias.matrix);
		clm_matrixSigmoid(outputs[b]);
	}
}

void clm_linearBackprop(clm_Linear *linear, float learnRate, unsigned int batchSize, clm_Matrix *inputs, bool updateErrors, clm_Matrix *outputErrors) {
	clm_Matrix *outputs = linear->output.matrixes;
	clm_Matrix *outputGradients = linear->gradient.matrixes;
	clm_Matrix *outputWeightsErrors = linear->weightsError.matrixes;
	clm_Matrix *inputErrors = linear->error.matrixes;

	for(unsigned int b = 0; b < batchSize; b++) {
		// clm_Matrix *inputsToThisLayer = i == 0 ? batchInputs : nn.layers[i - 1].output;
		// clm_Matrix *outputsOfThisLayer = nn.layers[i].output;
		//  clm_Matrix prevError = i == nn.numLayers - 1 ? INVALID_MATRIX : nn.layers[i + 1].error;
		//  clm_Matrix error = linear.error;

		// GPU step 1
		clm_matrixDSigmoid(clm_matrixCopy(outputs[b], outputGradients[b])); // dsig(yhat)
		clm_matrixMultiplyMatrixElements(outputGradients[b], inputErrors[b]); // (yhat - y) . dsig(yhat)
		clm_matrixMultiplyScalar(outputGradients[b], learnRate);

		// GPU Step 2
		clm_Matrix inputT = clm_matrixTranspose(inputs[b]);
		clm_matrixMultiplyMatrix(outputGradients[b], inputT, outputWeightsErrors[b]);

		// clm_matrixAddMatrix(linear->weights, outputWeightsErrors[b]);
		// clm_matrixAddMatrix(linear->bias, gradient);

		if(updateErrors) {
			// GPU Step 2 extended (opt)
			clm_Matrix weightsT = clm_matrixTranspose(linear->weights.matrix);
			clm_matrixMultiplyMatrix(weightsT, inputErrors[b], outputErrors[b]);
		}
	}
}
